function [Pareto,infeasibleSol,perfMetrics,algoQuota,tElapsed]=GALAXY(problem,popSize,NFEs,varargin)
% This is the master routine of the GALAXY method, which is a new hybrid
% optimiser for solving the discrete, multi-objective optimisation problems.

% --------------------------------------------------------------------------------------- %
%                                                                                         %
%	GGGGGGGGGGG	        A         LL                  A         XX     XX  YY      YY     %
%	GG				   AAA        LL                 AAA         XX   XX    YY    YY      %
%	GG				  AA AA       LL                AA AA         XX XX      YY  YY       %
%	GG				 AA   AA      LL               AA   AA         XXX        YYYY        %
%	GG	GGGGGGG	    AAAAAAAAA     LL              AAAAAAAAA        XXX         YY         %
%	GG		 GG	   AA       AA    LL             AA       AA      XX XX        YY         %
%	GG		 GG	  AA         AA   LL            AA         AA    XX   XX       YY         %
%	GGGGGGGGGGG	 AA           AA  LLLLLLLLLLL  AA           AA  XX     XX      YY         %
%                                                                                         %
% --------------------------------------------------------------------------------------- %

% --------------------------------------------------------------------------------------- %
% Input Vars:                                                                             %
%                                                                                         %
%   (1) problem: a valid name or index for the problem to be solved;                      %
%	(2) popSize: population size;                                                         %
%	(3) NFEs: number of function evaluations;                                             %
%   (4) varargin: variable length input argument list (Matlab built-in name);             %
%       optional for the basic settings of GALAXY;                                        %
%                                                                                         %
% Output Vars:                                                                            %
%                                                                                         %
%   (1) Pareto: Pareto set & Pareto front obtained; data is stored as follows:            %
%       [Pareto set | Pareto front | constrs | rank | crowding distance];                 %
%   (2) infeasibleSol: number of infeasible solutions in each population;                 %
%	(3)	perfMetrics: value(s) of performance metrice(s) if a reference front              %
%       is available for a specific problem;                                              %               
%   (4) algoQuota: variations of search operators;                                        %
%   (5) tElapsed: time elapsed;                                                           %
%                                                                                         %
% Example Usage:                                                                          %
%                                                                                         %
% [Pareto,infeasibleSol,perfMetrics,algoQuota,tElapsed]=GALAXY(2,100,25000,options);      %
% [Pareto,infeasibleSol,perfMetrics,algoQuota,tElapsed]=GALAXY('TLN',100,25000,options);  %
% [Pareto,infeasibleSol,perfMetrics,algoQuota,tElapsed]=GALAXY('TLN',100,25000);          %
% --------------------------------------------------------------------------------------- %

% --------------------------------------------------------------------------------------- %
% This algorithm has been presented at CCWI 2016:                                         %
%                                                                                         %
% Qi Wang, Dragan A. Savic, & Zoran Kapelan. GALAXY: A new hybrid optimizer for the       %
%     multi-objective design of water distribution systems, in Proceedings of 14th        %
%     CCWI Conference, Amsterdam, The Netherlands, 7-9 November, 2016.                    %
%                                                                                         %
% The GALAXY code was developed by Qi Wang, Guangdong University of Technology            % 
%     Email: wangqiguangzhou@163.com or q.wang@gdut.edu.cn                                %
%                                                                                         %
% Version 1.0:     December 2013    Cleaned up source code                                %
% Version 1.0.1:   December 2016    Bugs fixed for many-objective optimisation            %
% --------------------------------------------------------------------------------------- %

% --------------------------------------------------------------------------------------- %
%                                                                                         %
%               PLEASE CHECK THE TUTORIAL OF GALAXY FOR MORE DETAIL                       % 
%                                                                                         %
% Qi Wang, Dragan A. Savic, & Zoran Kapelan. A Brief Tutorial for the GALAXY Method.      %
%     Guangdong University of Technology, 20th January, 2017.                             %
%     It can be found in "...\galaxy-v1.0.1\document\GALAXY Tutorial_v1.0.1.pdf".         %
%                                                                                         %
% --------------------------------------------------------------------------------------- %

% profile on;
tStart=tic;
% =====PREPROCESSING=====
if isempty(varargin) % implement GALAXY using default settings
    options=struct('InitialisationMethod','INT',...
                            'SearchOperators',{{'TF','DE','SBXI','UM','GM','DC'}},...
                            'RangeOfMutation',0.7,...
                            'UseMex','Yes',...
                            'RecordPerformance','No',...
                            'PerformanceIndicators',{{'GD','HV','EI','EP'}},...
                            'PlotDuringOptimisation','No',...
                            'ScatterSize',10,...
                            'ScatterColor','blue');
else
    options=varargin{1};
end
% =====INITIALISATION=====
% retrieve the definition of benchmark WDS design problem
[problemDef]=problemArchive(problem);
% define some local variables
currentNFE=0;
numOfOperator=length(options.SearchOperators);
parents=zeros(popSize,problemDef.NumOfVar+problemDef.NumOfObj+3);
offspring=zeros(popSize,problemDef.NumOfVar+problemDef.NumOfObj+3);
approxiFront=zeros(popSize,problemDef.NumOfObj);
infeasibleSol=zeros(NFEs/popSize,1);
vars=zeros(popSize,problemDef.NumOfVar);
% initialise the variables
switch options.InitialisationMethod
    case 'LHS' % Latin hypercube sampling from within the ranges of variables
        vars=LHSI(problemDef.VarLowerBound,problemDef.VarUpperBound,popSize);
    case 'EXT' % load the initial population from an external file
        initialPopulationFileName=varargin{2};
        vars=floor(load(initialPopulationFileName));
    case 'INT' % uniformly distributed random integers within the ranges of variables
        for n=1:problemDef.NumOfVar
            vars(:,n)=randi([problemDef.VarLowerBound(n) problemDef.VarUpperBound(n)],popSize,1);
        end
    otherwise
        disp('WARNING: The initialisation method you specified does not exist. Please check carefully.');
        Pareto=[];
        infeasibleSol=[];
        perfMetrics=[];
        algoQuota=[];
        tElapsed=[];
        return;
end
% initialise the quota of each search operator
remainingSlots=mod(popSize,numOfOperator);
if remainingSlots==0
    quota=(popSize/numOfOperator)*ones(1,numOfOperator);
else
    quota=floor(popSize/numOfOperator)*ones(1,numOfOperator);
    quota(1)=quota(1)+remainingSlots; % give the operator with the best leaping ability in the global sense as many slots as possible
end
% =====EVALUATION=====
[objs,constrs]=objectiveFunction(vars,NFEs,problemDef,options);
currentNFE=currentNFE+popSize;
currentGen=currentNFE/popSize;
infeasibleSol(currentGen)=sum(constrs~=0);
% sort the initial population
[ranking,crowdDist]=ConstrainedSorting(objs,constrs);
% =====PERFORMANCE=====
% evaluate the quality of initial population if required
if strcmpi(options.RecordPerformance,'yes')
    perfMetrics=zeros(NFEs/popSize,4);
    GD=0;
    HV=0;
    EI=0;
    EP=0;
    approxiFront=objs;
    referenceFront=problemDef.PF;
    maxCost=problemDef.MaxCost;
    epsilon=problemDef.Epsilon;
    [GD,HV_RF]=performanceEval(options.PerformanceIndicators{1},approxiFront,referenceFront,constrs,ranking,[],maxCost,epsilon);
    [HV,HV_RF]=performanceEval(options.PerformanceIndicators{2},approxiFront,referenceFront,constrs,ranking,[],maxCost,epsilon);
    [EI,HV_RF]=performanceEval(options.PerformanceIndicators{3},approxiFront,referenceFront,constrs,ranking,[],maxCost,epsilon);
    [EP,HV_RF]=performanceEval(options.PerformanceIndicators{4},approxiFront,referenceFront,constrs,ranking,[],maxCost,epsilon);
    perfMetrics(currentGen,1)=GD;
    perfMetrics(currentGen,2)=HV;
    perfMetrics(currentGen,3)=EI;
    perfMetrics(currentGen,4)=EP;
else
    perfMetrics=[];
end
algoQuota(currentGen,:)=quota;
algoEffic(currentGen,1)=0;
% =====VISUALISATON=====
if strcmpi(options.PlotDuringOptimisation,'yes')
    figureHandle=[];
    figureHandle=MOEAVisualisation(approxiFront,referenceFront,problemDef.Name,problemDef.Index,currentGen,figureHandle,options);
end
% =====MAIN LOOP=====
parents=[vars,objs,constrs,ranking,crowdDist];
while currentNFE<NFEs
    % =====SELECTION=====
    % A very simple selection procedure is implemented here, i.e., select all the individuals as parents.
    % Thus, it is the REPLACEMENT procedure that steers the optimisation towards the Pareto-optimal front.
    % =====GENERATION=====
    individuals=[];
    offspringIdentity=zeros(popSize,1);
    for i=1:numOfOperator
        [vars]=generateOffspring(parents,options.SearchOperators{i},quota(i),problemDef,options);
        individuals=[individuals;vars];
        offspringIdentity=[offspringIdentity;i*ones(quota(i),1)];
    end
    % check whether variables are in bound
    lowerBound=repmat(problemDef.VarLowerBound,popSize,1);
    upperBound=repmat(problemDef.VarUpperBound,popSize,1);
    individuals(individuals<lowerBound)=lowerBound(individuals<lowerBound);
    individuals(individuals>upperBound)=upperBound(individuals>upperBound);
    % =====EVALUATION=====
    [objs,constrs]=objectiveFunction(individuals,NFEs,problemDef,options);
    offspring=[individuals,objs,constrs, zeros(popSize,2)];
    % =====REPLACEMENT=====
    [parents,offspringIdentity]=HybridReplacement(parents,offspring,offspringIdentity,problemDef);
    % =====UPDATE=====
    [quota,efficiency]=UpdateQuota(quota,offspringIdentity,options);
    currentNFE=currentNFE+popSize;
    currentGen=currentNFE/popSize;
    infeasibleSol(currentGen)=sum(parents(:,end-2)~=0);
    % =====PERFORMANCE=====
    % evaluate the quality of current population if required
    if strcmpi(options.RecordPerformance,'yes')
        approxiFront=parents(:,problemDef.NumOfVar+1:problemDef.NumOfVar+problemDef.NumOfObj);
        constrs=parents(:,end-2);
        ranking=parents(:,end-1);
        [GD,HV_RF]=performanceEval(options.PerformanceIndicators{1},approxiFront,referenceFront,constrs,ranking,HV_RF,maxCost,epsilon);
        [HV,HV_RF]=performanceEval(options.PerformanceIndicators{2},approxiFront,referenceFront,constrs,ranking,HV_RF,maxCost,epsilon);
        [EI,HV_RF]=performanceEval(options.PerformanceIndicators{3},approxiFront,referenceFront,constrs,ranking,HV_RF,maxCost,epsilon);
        [EP,HV_RF]=performanceEval(options.PerformanceIndicators{4},approxiFront,referenceFront,constrs,ranking,HV_RF,maxCost,epsilon);
        perfMetrics(currentGen,1)=GD;
        perfMetrics(currentGen,2)=HV;
        perfMetrics(currentGen,3)=EI;
        perfMetrics(currentGen,4)=EP;
    else
        perfMetrics=[];
    end
    algoQuota(currentGen,:)=quota;
    algoEffic(currentGen,1)=efficiency;
    % =====VISUALISATON=====
    if strcmpi(options.PlotDuringOptimisation,'yes')
        figureHandle=MOEAVisualisation(approxiFront,referenceFront,problemDef.Name,problemDef.Index,currentGen,figureHandle,options);
    end
    disp(['Current NFEs is ' num2str(currentNFE)]);
end
% =====OUTPUT=====
Pareto=parents;
tElapsed=toc(tStart);
disp(['Total time elapsed is ' num2str(tElapsed) ' seconds.']);
% profile viewer;

function vars=LHSI(lowerBound,upperBound,popSize)
% This function is adapted from the 'LHSU' function within the source code
% of AMALGAM written by Jasper A. Vrugt. It uses the Latin hypercube
% sampling technique to initialise the population with respect to the
% range of each decision variable.

% define local variables
numOfVar=length(lowerBound);
vars=rand(popSize,numOfVar);
upperBound=upperBound+1;
for i=1:numOfVar,
    idx=randperm(popSize);
    P=(idx'-vars(:,i))/popSize;
    vars(:,i)=lowerBound(i)+P.*(upperBound(i)-lowerBound(i));
end
% convert from real to integer
vars=floor(vars);

function [offspring]=generateOffspring(parent,searchOperator,quota,problemDef,options)
% This function produces a predefined number of offspring using the specified search operator.

% define local variables
popSize=size(parent,1);
numOfVar=problemDef.NumOfVar;
numOfObj=problemDef.NumOfObj;
offspring=zeros(popSize,numOfVar);
candidateList1=zeros(1,popSize);
candidateList2=zeros(1,popSize);
individual1=0;
individual2=0;
individual3=0;
individual4=0;
parent1=zeros(1,numOfVar);
parent2=zeros(1,numOfVar);
pM=1/numOfVar;
mutLoc=zeros(popSize,numOfVar);
unifRndInt=zeros(popSize,numOfVar);
norm_mean=zeros(1,numOfVar);
norm_std=zeros(1,numOfVar);
normRndInt=zeros(popSize,numOfVar);
P_dcm=0;
P_dcm_min=0;
P_dcm_max=0;
switch searchOperator
    case 'TF' % Turbulence Factor
        for i=1:popSize
            offspring(i,:)=floor(parent(i,1:numOfVar)+unifrnd(-1,1,1,1).*parent(i,1:numOfVar));
        end
    case 'DE' % Differential Evolution (for integer coding scheme)
        for i=1:popSize
            candidateList1=randperm(popSize);
            individual1=candidateList1(1);
            individual2=candidateList1(2);
            individual3=candidateList1(3);
            offspring(i,:)=parent(i,1:numOfVar)+(parent(individual3,1:numOfVar)-parent(i,1:numOfVar))+(parent(individual1,1:numOfVar)-parent(individual2,1:numOfVar));
        end
    case 'SBXI' % Simulated Binary Crossover for Integers
        offspring=[];
        pC=1.0;
        candidateList1=randperm(popSize);
        candidateList2=randperm(popSize);
        for i=0:4:popSize-4
            % -----First Round-----
            % select parents
            individual1=candidateList1(i+1);
            individual2=candidateList1(i+2);
            individual3=candidateList1(i+3);
            individual4=candidateList1(i+4);
            % The tournament part has been corrected on 28-Nov-2016 [VERY IMPORTANT]. 
            parent1=TournamentSelection(parent(individual1,1:numOfVar),parent(individual2,1:numOfVar),...
                parent(individual1,end-2),parent(individual2,end-2),...
                parent(individual1,numOfVar+1:numOfVar+numOfObj),parent(individual2,numOfVar+1:numOfVar+numOfObj),...
                parent(individual1,end),parent(individual2,end),...
                numOfObj);
            parent2=TournamentSelection(parent(individual3,1:numOfVar),parent(individual4,1:numOfVar),...
                parent(individual3,end-2),parent(individual4,end-2),...
                parent(individual3,numOfVar+1:numOfVar+numOfObj),parent(individual4,numOfVar+1:numOfVar+numOfObj),...
                parent(individual3,end),parent(individual4,end),...
                numOfObj);
            % crossover for integers
            [child1,child2]=CrossoverInteger(parent1,parent2,pC);
            % append the children to the offspring
            offspring=[offspring;child1;child2];
            % -----Second Round-----
            % select parents
            individual1=candidateList2(i+1);
            individual2=candidateList2(i+2);
            individual3=candidateList2(i+3);
            individual4=candidateList2(i+4);
            % The tournament part has been corrected on 28-Nov-2016 [VERY IMPORTANT]. 
            parent1=TournamentSelection(parent(individual1,1:numOfVar),parent(individual2,1:numOfVar),...
                parent(individual1,end-2),parent(individual2,end-2),...
                parent(individual1,numOfVar+1:numOfVar+numOfObj),parent(individual2,numOfVar+1:numOfVar+numOfObj),...
                parent(individual1,end),parent(individual2,end),...
                numOfObj);
            parent2=TournamentSelection(parent(individual3,1:numOfVar),parent(individual4,1:numOfVar),...
                parent(individual3,end-2),parent(individual4,end-2),...
                parent(individual3,numOfVar+1:numOfVar+numOfObj),parent(individual4,numOfVar+1:numOfVar+numOfObj),...
                parent(individual3,end),parent(individual4,end),...
                numOfObj);
            % crossover for integers
            [child1,child2]=CrossoverInteger(parent1,parent2,pC);
            % append the children to the offspring
            offspring=[offspring;child1;child2];
        end;
        % check the size of offspring, should be equal to popSize
        if size(offspring,1)<popSize % possible to lack 2 children if the population size cannot be divided by 4
            individual1=candidateList1(end-1);
            individual2=candidateList1(end);
            individual3=candidateList1(round(0.5*popSize));
            individual4=candidateList1(round(0.5*popSize));
            % select parents
            parent1=TournamentSelection(parent(individual1,1:numOfVar),parent(individual2,1:numOfVar),...
                parent(individual1,end-2),parent(individual2,end-2),...
                parent(individual1,numOfVar+1:numOfVar+numOfObj),parent(individual2,numOfVar+1:numOfVar+numOfObj),...
                parent(individual1,end),parent(individual2,end),...
                numOfObj);
            parent2=TournamentSelection(parent(individual3,1:numOfVar),parent(individual4,1:numOfVar),...
                parent(individual3,end-2),parent(individual4,end-2),...
                parent(individual3,numOfVar+1:numOfVar+numOfObj),parent(individual4,numOfVar+1:numOfVar+numOfObj),...
                parent(individual3,end),parent(individual4,end),...
                numOfObj);
            % crossover for integers
            [child1,child2]=CrossoverInteger(parent1,parent2,pC);
            % append the children to the offspring
            offspring=[offspring;child1;child2];
        end;
    case 'UM' % Uniform Mutation
        offspring=parent(:,1:numOfVar);
        mutLoc=rand(popSize,numOfVar)<=pM;
        for i=1:numOfVar
            unifRndInt(:,i)=randi([problemDef.VarLowerBound(i),problemDef.VarUpperBound(i)],popSize,1);
        end
        offspring(mutLoc)=unifRndInt(mutLoc);
    case 'GM' % Gaussian Mutation
        offspring=parent(:,1:numOfVar);
        mutLoc=rand(popSize,numOfVar)<=pM;
        % use 10 sigma to control the range of normal distribution
        norm_mean=mean([problemDef.VarLowerBound;problemDef.VarUpperBound]);
        norm_std=(problemDef.VarUpperBound-problemDef.VarLowerBound+1)/10;
        normRndInt=floor(repmat(norm_mean,popSize,1)+repmat(norm_std,popSize,1).*randn(popSize,numOfVar));
        offspring(mutLoc)=normRndInt(mutLoc);
    case 'DC' % Dither Creeping
        ratio=options.RangeOfMutation;
        P_dcm_min=(1-ratio)*pM;
        P_dcm_max=(1+ratio)*pM;
        P_d=rand(1);
        offspring=parent(:,1:numOfVar);
        for i=1:popSize
            for j=1:numOfVar
                P_dcm=P_dcm_min+rand(1)*(P_dcm_max-P_dcm_min);
                if rand(1)<P_dcm
                    if rand(1)<=P_d
                        offspring(i,j)=max(problemDef.VarLowerBound(j),parent(i,j)-1);
                    else
                        offspring(i,j)=min(problemDef.VarUpperBound(j),parent(i,j)+1);
                    end
                end
            end
        end
end
% randomly select a portion of individuals (equal to quota) from the offspring
candidateList1=randperm(popSize);
offspring=offspring(candidateList1(1:quota),:);

function [parent]=TournamentSelection(vars1,vars2,constrs1,constrs2,objs1,objs2,crowdDist1,crowdDist2,numOfObj)
% This function implements the tournament selection.

% define local variables
selected=0;
% check Pareto-dominance relations
if constrs1==0 && constrs2==0 % both are feasible solutions
    for i=1:numOfObj
        dom_less=0;
        dom_more=0;
        dom_equal=0;
        if objs1(i)<objs2(i)
            dom_less=dom_less+1;
        elseif objs1(i)>objs2(i)
            dom_more=dom_more+1;
        else
            dom_equal=dom_equal+1;
        end
    end
    if dom_more==0 && dom_equal~=numOfObj
        parent=vars1;
        selected=1;
    elseif dom_less==0 && dom_equal~=numOfObj
        parent=vars2;
        selected=1;
    else
        selected=0;
    end
else % at least one of them is infeasible
    if constrs1>constrs2
        parent=vars1;
        selected=1;
    elseif constrs1<constrs2
        parent=vars2;
        selected=1;
    else
        selected=0;
    end
end
% use the crowding distance to distinguish two solutions if the tie still exists
if selected==0
    if crowdDist1>crowdDist2
        parent=vars1;
        selected=1;
    elseif crowdDist1<crowdDist2
        parent=vars2;
        selected=1;
    else
        if rand<=0.5
            parent=vars1;
            selected=1;
        else
            parent=vars2;
            selected=1;
        end
    end
end

function [child1,child2]=CrossoverInteger(parent1,parent2,pC)
% This function implements the simulated binary crossover for integers.

% define local variables
numOfVar=size(parent1,2);
child1=zeros(size(parent1));
child2=zeros(size(parent2));
% simulated binary crossover for integers
if rand(1)<=pC
    for i=1:numOfVar
        if rand(1)<=0.5
            if abs(parent1(i)-parent2(i))~=0
                if parent1(i)<parent2(i)
                    y1=parent1(i); y2=parent2(i);
                else
                    y1=parent2(i); y2=parent1(i);
                end
                if rand(1)<=0.5
                    child1(i)=y1+randi(y2-y1,1);
                    child2(i)=y1+randi(y2-y1,1);
                else
                    child1(i)=y2-randi(y2-y1,1);
                    child2(i)=y2-randi(y2-y1,1);
                end
            else
                child1(i)=parent1(i);
                child2(i)=parent2(i);
            end
        else
            child1(i)=parent1(i);
            child2(i)=parent2(i);
        end
    end
else
    child1=parent1;
    child2=parent2;
end

function [parent,offspringIdentity]=HybridReplacement(parent,offspring,offspringIdentity,problemDef)
% This function implements the hybrid replacement strategy proposed for GALAXY.
% The hybrid replacement strategy is based on the traditional
% Pareto-dominance and the epsilon-dominance concepts but does not require any
% pre-specified epsilon precisions. It is applied when the number of
% solutions in the first rank exceeds the population size. The idea here is
% to prevent the solutions on the Pareto-optimal front from being removed
% by the crowding distance based selection procedure.

% define local variables
rank=1;
numOfOffspring=size(parent,1);
numOfVar=problemDef.NumOfVar;
numOfObj=problemDef.NumOfObj;
tempPop=zeros(2*numOfOffspring,numOfVar);
tempObjs=zeros(2*numOfOffspring,numOfObj);
tempConstrs=zeros(2*numOfOffspring,1);
% append children to parents
tempPop=[parent(:,1:numOfVar);offspring(:,1:numOfVar)];
tempObjs=[parent(:,numOfVar+1:numOfVar+numOfObj);offspring(:,numOfVar+1:numOfVar+numOfObj)];
tempConstrs=[parent(:,end-2);offspring(:,end-2)];
% remove duplicates
[tempPop,idxUnique]=unique(tempPop,'rows');
tempObjs=tempObjs(idxUnique,:);
tempConstrs=tempConstrs(idxUnique,:);
offspringIdentity=offspringIdentity(idxUnique,:);
% rank parent and offspring
[Ranking,CrowdDist]=ConstrainedSorting(tempObjs,tempConstrs);
% generate new population
for i=1:max(Ranking)
    countInThisRank=length(find(Ranking==i));
    countRank(i)=countInThisRank;
end
sumCountRank=cumsum(countRank);
if sumCountRank(1)==numOfOffspring % select all the individuals in the first rank
    idx=find(Ranking==1);
    selectedIndex=idx;
elseif sumCountRank(1)>numOfOffspring
    idxPreserved=[];
    % find the boundary solutions
    idxFirstPF=find(Ranking==1);
    [sortedFirstPF,~]=sortrows(tempObjs(idxFirstPF,:),1);
    idxLB=ismember(tempObjs(idxFirstPF,:),sortedFirstPF(1,:),'rows');
    idxUB=ismember(tempObjs(idxFirstPF,:),sortedFirstPF(end,:),'rows');    
    idxPreserved=[idxFirstPF(idxLB);idxFirstPF(idxUB)];
    remainingSlots=numOfOffspring-length(idxPreserved);
    idxCandidate=setdiff(idxFirstPF,idxPreserved);
    % construct the grids with epsilon precision
    gridLB=min(tempObjs(idxFirstPF,:));
    gridUB=max(tempObjs(idxFirstPF,:));
    epsilon=(gridUB-gridLB)/(remainingSlots);
    idenArray=floor((tempObjs(idxCandidate,:)-repmat(gridLB,size(tempObjs(idxCandidate,:),1),1))./repmat(epsilon,size(tempObjs(idxCandidate,:),1),1));
    % pick the epsilon-dominating solutions
    idxSelected=[];
    idxUnselected=[];
    numOfEmptyDim=0;
    for i=1:remainingSlots
        idxSameDim=idxCandidate(idenArray(:,1)==i-1);
        if isempty(idxSameDim)
            % record this empty column
            numOfEmptyDim=numOfEmptyDim+1;
        else
            slicedDim=idenArray(idenArray(:,1)==i-1,:);
            [sortedDim,idxSortedDim]=sort(sum(idenArray(idenArray(:,1)==i-1,2:end),2));
            coorBestGrid=slicedDim(idxSortedDim(1),:);
            idxSameGrid=ismember(slicedDim,coorBestGrid,'rows');
            idxNumOfSol=find(idxSameGrid==1);
            idxEpsDominated=find(idxSameGrid~=1);
            idxUnselected=[idxUnselected;idxSameDim(idxEpsDominated)];
            if length(idxNumOfSol)==1
                idxSelected=[idxSelected; idxSameDim(idxNumOfSol)];
            elseif length(idxNumOfSol)>1
                dist=sqrt(sum((tempObjs(idxSameDim(idxNumOfSol),:)-repmat(coorBestGrid,size(tempObjs(idxSameDim(idxNumOfSol),:),1),1)).^2,2));
                [~,idxMinDist]=min(dist);
                idxSelected=[idxSelected;idxSameDim(idxNumOfSol(idxMinDist))];
                idxUnselected=[idxUnselected;setdiff(idxSameDim(idxNumOfSol),idxSelected)];
            end
        end
    end
    % it is possible that some not-so-good solutions have to be inserted
    dist=sqrt(sum((tempObjs(idxUnselected,:)-repmat(zeros(1,numOfObj),size(tempObjs(idxUnselected,:),1),1)).^2,2));
    [~,idxSortedDist]=sort(dist);
    idxSelected=[idxSelected; idxUnselected(idxSortedDist(1:numOfEmptyDim))];
    idxPreserved=[idxPreserved;idxSelected];
    selectedIndex=idxPreserved;
elseif sumCountRank(1)<numOfOffspring
    % check which rank contains the number of solutions that are within the size of the population
    idx=find(sumCountRank>numOfOffspring);
    rk=idx(1)-1;
    if sumCountRank(rk)~=numOfOffspring,
        % fill population by selecting from next population
        idx=find(Ranking==rk+1);
        % check how many individuals are missing
        T=numOfOffspring-sumCountRank(rk);
        % select based on the descending order of crowding distance
        [~,sortedIdx]=sort(CrowdDist(idx),'descend');
        idx=idx(sortedIdx);
        selectedIndex=idx(1:T);
    else
        selectedIndex=[];
    end;
    % now fill with all the other ranks
    selectedIndex=[selectedIndex;find(Ranking<=rk)];
end
% generate new population
vars=tempPop(selectedIndex,:);
objs=tempObjs(selectedIndex,:);
constrs=tempConstrs(selectedIndex,:);
parent=[vars,objs,constrs,Ranking(selectedIndex,:),CrowdDist(selectedIndex,:)];
% update offspringIdentity
offspringIdentity=offspringIdentity(selectedIndex,:);

function [ranking,crowdDist]=ConstrainedSorting(objs,constrs)
% This function implements the non-dominated sorting and crowding distance sorting proposed in Deb et al. (2002).

% define local variables
[numOfSol,numOfObj]=size(objs);
ranking=zeros(size(objs,1),1);
crowdDist=zeros(size(objs,1),1);
front=1;
rank(front).sol=[];
solution=struct('countOfDominated',0,'setOfDominating',[]);
% find the solutions in the first rank
for i=1:numOfSol
    solution(i).countOfDominated=0;
    solution(i).setOfDominating=[];
    for j=1:numOfSol
        if constrs(i)==0 && constrs(j)==0 % both are feasible solutions
            dom_less=0;
            dom_equal=0;
            dom_more=0;
            for k=1:numOfObj
                if (objs(i,k)<objs(j,k))
                    dom_less=dom_less+1;
                elseif (objs(i,k)==objs(j,k))
                    dom_equal=dom_equal+1;
                else
                    dom_more=dom_more+1;
                end
            end
            if dom_less==0 && dom_equal~=numOfObj
                solution(i).countOfDominated=solution(i).countOfDominated+1;
            elseif dom_more==0 && dom_equal~=numOfObj
                solution(i).setOfDominating=[solution(i).setOfDominating,j];
            end
        elseif constrs(i)>constrs(j)
            solution(i).setOfDominating=[solution(i).setOfDominating,j];
        elseif constrs(i)<constrs(j)
            solution(i).countOfDominated=solution(i).countOfDominated+1;
        end
    end
    if solution(i).countOfDominated==0
        ranking(i)=1;
        rank(front).sol=[rank(front).sol,i];
    end
end
% fine the solutions in the remaining ranks
while ~isempty(rank(front).sol)
    currentRankSol=[];
    for i=1:length(rank(front).sol)
        if ~isempty(solution(rank(front).sol(i)).setOfDominating)
            for j=1:length(solution(rank(front).sol(i)).setOfDominating)
                solution(solution(rank(front).sol(i)).setOfDominating(j)).countOfDominated=solution(solution(rank(front).sol(i)).setOfDominating(j)).countOfDominated-1;
                if solution(solution(rank(front).sol(i)).setOfDominating(j)).countOfDominated==0
                    ranking(solution(rank(front).sol(i)).setOfDominating(j))=front+1;
                    currentRankSol=[currentRankSol,solution(rank(front).sol(i)).setOfDominating(j)];
                end
            end
        end
    end
    front=front+1;
    rank(front).sol=currentRankSol;
end
currentIdx=0;
% compute the crowding distance for each individual in each front
for front=1:(front-1)
    previousIdx=currentIdx+1;
    currentIdx=currentIdx+length(rank(front).sol);
    dist=zeros(currentIdx-previousIdx+1,numOfObj);
    tmpObjs=objs(ranking==front,:);
    for i=1:numOfObj
        [~,idxObj]=sort(tmpObjs(:,i));
        objMin=tmpObjs(idxObj(1),i);
        objMax=tmpObjs(idxObj(end),i);
        dist(idxObj(1),i)=Inf;
        dist(idxObj(end),i)=Inf;
        for j=2:length(idxObj)-1
            nextObj=tmpObjs(idxObj(j+1),i);
            previousObj=tmpObjs(idxObj(j-1),i);
            if (objMax-objMin==0)
                dist(idxObj(j),i)=Inf;
            else
                dist(idxObj(j),i)=(nextObj-previousObj)/(objMax-objMin);
            end
        end
    end
    distance=zeros(length(rank(front).sol),1);
    for i=1:numOfObj
        distance(:,1)=distance(:,1)+dist(:,i);
    end
    crowdDist(previousIdx:currentIdx,1)=distance;
end
[~,idx]=sort(ranking);
crowdDist(idx,:)=crowdDist;

function [newQuota,efficiency]=UpdateQuota(oldQuota,offspringIdentity,options)
% This function calculates the contribution of each search operator and update its quota for the next generation.

% define local variables
numOfSol=size(offspringIdentity,1);
numOfSurvival=sum(offspringIdentity>0);
efficiency=numOfSurvival/size(offspringIdentity,1);
numOfOperator=length(options.SearchOperators);
percentage=zeros(1,numOfOperator);
member=zeros(1,numOfOperator);
% calculate the contribution of each search operator in percentage term
for i=1:numOfOperator
    percentage(1,i)=sum(offspringIdentity==i)/oldQuota(i);
end
percentageSurvival=sum(percentage);
% update the quota of each search operator
if percentageSurvival==0 % reallocate the slots when all operators failed
    if mod(numOfSol,numOfOperator)==0
        newQuota=(numOfSol/numOfOperator)*ones(1,numOfOperator);
    else
        newQuota=floor(numOfSol/numOfOperator)*ones(1,numOfOperator);
        remainingQuota=numOfSol-numOfOperator*floor(numOfSol/numOfOperator);
        for i=1:remainingQuota
            newQuota(i)=newQuota(i)+1;
        end
    end
else
    scaling=1/percentageSurvival;
    member=max(floor(scaling*percentage*numOfSol),0);
    diff=(numOfSol-sum(member));
    % prefer the best operator by giving slots as many as possible
    idxMax=find(member==max(member));
    member(idxMax(1))=member(idxMax(1))+diff;
    % keep the balance of multiple operators
    if min(member)==0
        [idxMin]=find(member==0);
        [~,order]=sort(member);
        member(idxMin)=1; % ensure that the worst operator can still produce at least one offspring
        numOfNonWorstOperator=numOfOperator-length(idxMin);
        % always borrowing slots from the topmost operators
        if numOfNonWorstOperator==1
            member(order(end))=member(order(end))-1*length(idxMin);
        elseif numOfNonWorstOperator==2 && length(idxMin)==1
            member(order(end))=member(order(end))-1;
        elseif numOfNonWorstOperator==2 && length(idxMin)>1
            member(order(end-1))=member(order(end-1))-1;
            member(order(end))=member(order(end))-1*(length(idxMin)-1);
        elseif numOfNonWorstOperator==3 && length(idxMin)==1
            member(order(end))=member(order(end))-1;
        elseif numOfNonWorstOperator==3 && length(idxMin)==2
            member(order(end))=member(order(end))-1;
            member(order(end-1))=member(order(end-1))-1;
        elseif numOfNonWorstOperator==3 && length(idxMin)==3
            member(order(end))=member(order(end))-1;
            member(order(end-1))=member(order(end-1))-1;
            member(order(end-2))=member(order(end-2))-1;
        elseif numOfNonWorstOperator==4 && length(idxMin)==1
            member(order(end))=member(order(end))-1;
        elseif numOfNonWorstOperator==4 && length(idxMin)==2
            member(order(end))=member(order(end))-1;
            member(order(end-1))=member(order(end-1))-1;
        elseif numOfNonWorstOperator==5 && length(idxMin)==1
            member(order(end))=member(order(end))-1;
        end
    end
    newQuota=member;
end