function [indicatorValue,HV_RF]=performanceEval(indicatorName,approxiFront,referenceFront,varargin)
% This function calculates the quality of approximation front according to a specified performance indicator.

% The performance indicators currently supported in this function include:
% (1) Generational Distance (GD) (Van Veldhuizen & Lamont, 1998): measure an average distance of the non-dominated solutions in the approximation set from a reference set
% (2) Hypervolume (HV) (Zitzler, 1999): calculate the ratio of volume of hypercube dominated by the approximation set to the volume dominated by the reference set
% (3) Additive Epsilon Indicator (EI) (Zitzler et al., 2003): compute the minimum distance it is required to convert the approximation front to dominate the reference front
% (4) Epsilon Performance (EP) (Kollat & Reed, 2005): check whether the solutions found are within the constructed hypercube of each solution in the reference front

% References:
% Van Veldhuizen, D. A., & Lamont, G. B. (1998, July). Evolutionary computation and convergence to a pareto front. In Late breaking papers at the genetic programming 1998 conference (pp. 221-228). Stanford University Bookstore, University of Wisconsin, Madison, Wisconsin, USA.
% Zitzler, E. (1999). Evolutionary algorithms for multiobjective optimization: Methods and applications (Vol. 63). Ithaca: Shaker.
% Zitzler, E., Thiele, L., Laumanns, M., Fonseca, C. M., & Da Fonseca, V. G. (2003). Performance assessment of multiobjective optimizers: An analysis and review. Evolutionary Computation, IEEE Transactions on, 7(2), 117-132.
% Kollat, J. B., & Reed, P. M. (2005, January). The value of online adaptive search: a performance comparison of NSGAII, ?-NSGAII and ?MOEA. In Evolutionary Multi-Criterion Optimization (pp. 386-398). Springer Berlin Heidelberg.

% Input Vars:
%	(1) indicatorName: acronym of a certain performance indicator;
%	(2) approxiFront: approximation front obtained by an algorithm;
%	(3) referenceFront: reference front loaded from hard disk;
%	(4) varargin: variable length input argument list (Matlab built-in name);
% Output Vars:
%	(1) indicatorValue: a vector of specified performance indicators;
%   (2) HV_RF: hypervolume of reference front (to avoid repetitive computation)

% data pre-processing (make sure that solutions in the approximation front are feasible with the topmost rank)
constrViol=varargin{1};
Ranking=varargin{2};
HV_RF=varargin{3}; % hypervolume of reference front
maxCost=varargin{4}; % unit: million monetary
epsilon=varargin{5}; % given in the format of [eps_In, eps_cost]
validIndex=constrViol(Ranking==1);
approxiFront=approxiFront(Ranking==1,:);
approxiFront=approxiFront(validIndex==0,:);
AF=approxiFront;
RF=referenceFront;
% compute the value of the specified performance indicator
switch indicatorName
    case 'GD'
        if isempty(approxiFront)
            indicatorValue=Inf;
        else
            AF=[abs(AF(:,2)),AF(:,1)]; % change the position of two columns and convert negative network resilience to positive value
            RF(:,2)=RF(:,2)/maxCost;
            AF(:,2)=AF(:,2)/maxCost;
            [nRF,~]=size(RF);
            [nAF,~]=size(AF);
            dist=zeros(nRF,1);
            minDist=zeros(nAF,1);
            for i=1:nAF
                dist=sqrt(sum((RF-repmat(AF(i,:),nRF,1)).^2,2));
                minDist(i)=min(dist); % for each solution in the RF, find the minimum distance from it to AF
            end
            GD=mean(minDist);
            indicatorValue=GD;
        end
    case 'HV'
        referencePoint=[1 0]; % referent point for the normalised axes denoting [max(Cost), min(In)]
        RF(:,2)=RF(:,2)/maxCost;
        if isempty(AF)
            indicatorValue=0;
            if isempty(HV_RF)
                sortedRF=sortrows(RF,1);
                boundaryCoordinateRF=[referencePoint; referencePoint(1)*ones(size(sortedRF,1)-1,1) sortedRF(1:end-1,1)];
                hvSortedRF=abs(sortedRF(:,2)-boundaryCoordinateRF(:,1)).*abs(sortedRF(:,1)-boundaryCoordinateRF(:,2));
                HV_RF=sum(hvSortedRF);
            end
        else
            AF=[abs(AF(:,2)),AF(:,1)]; % change the position of two columns, and convert negative network resilience to positive value
            AF(:,2)=AF(:,2)/maxCost;
            sortedAF=sortrows(AF,1);
            boundaryCoordinateAF=[referencePoint; referencePoint(1)*ones(size(sortedAF,1)-1,1) sortedAF(1:end-1,1)];
            hvSortedAF=abs(sortedAF(:,2)-boundaryCoordinateAF(:,1)).*abs(sortedAF(:,1)-boundaryCoordinateAF(:,2));
            HV_AF=sum(hvSortedAF);
            if isempty(HV_RF)
                sortedRF=sortrows(RF,1);
                boundaryCoordinateRF=[referencePoint; referencePoint(1)*ones(size(sortedRF,1)-1,1) sortedRF(1:end-1,1)];
                hvSortedRF=abs(sortedRF(:,2)-boundaryCoordinateRF(:,1)).*abs(sortedRF(:,1)-boundaryCoordinateRF(:,2));
                HV_RF=sum(hvSortedRF);
            end
            indicatorValue=HV_AF/HV_RF;
        end
    case 'EI'
        if isempty(AF)
            indicatorValue=Inf;
        else
            AF=[abs(AF(:,2)),AF(:,1)]; % change the position of two columns, and convert negative network resilience to positive value
            RF(:,2)=RF(:,2)/maxCost;
            AF(:,2)=AF(:,2)/maxCost;
            % first objective
            AF_1=AF(:,1);
            RF_1=RF(:,1);
            [mAF_1,mRF_1]=meshgrid(AF_1,RF_1);
            diff_1=mRF_1-mAF_1; % maximisation of network resilience
            % second objective
            AF_2=AF(:,2);
            RF_2=RF(:,2);
            [mAF_2,mRF_2]=meshgrid(AF_2,RF_2);
            diff_2=mAF_2-mRF_2; % minimisation of cost
            diff_comb=max(diff_1,diff_2);
            indicatorValue=max(min(diff_comb,[],2));
        end
    case 'EP'
        if isempty(AF)
            indicatorValue=0;
        else
            AF=[abs(AF(:,2)),AF(:,1)]; % change the position of two columns, and convert negative network resilience to positive value
            % first objective
            AF_1=AF(:,1);
            RF_1=RF(:,1);
            [mAF_1,mRF_1]=meshgrid(AF_1,RF_1);
            diff_1=abs(mRF_1-mAF_1);
            % second objective
            AF_2=AF(:,2);
            RF_2=RF(:,2);
            [mAF_2,mRF_2]=meshgrid(AF_2,RF_2);
            diff_2=abs(mAF_2-mRF_2);
            diff_sum=sqrt(diff_1.^2+diff_2.^2);
            [~, min_diff_idx]=min(diff_sum,[],2);
            nCount=0;
            for i=1:size(RF,1)
                if diff_1(i,min_diff_idx(i))<=0.5*epsilon(1) && diff_2(i,min_diff_idx(i))<=0.5*epsilon(2)
                    nCount=nCount+1;
                end
            end
            indicatorValue=nCount/size(RF,1);
        end
    otherwise
        disp('No such method exists. Please check carefully.');
        indicatorValue=[];
end